@@ -1,12 +1,14 @@
 extern crate rna_ss_params;
-extern crate bio_seq_algos;
 extern crate getopts;
 extern crate itertools;
 extern crate num;
 extern crate scoped_threadpool;
 extern crate bio;
 extern crate num_cpus;
+extern crate hashbrown;
 
 pub mod utils;
 pub mod mccaskill_algo;
 pub mod centroid_estimator;
+pub mod compiled_seq_align_params;
+pub mod durbin_algo;
